import cv2
import numpy as np
from numpy import sin, cos

def warp(im, T, out_im):
    ih, iw = im.shape[:2]
    oh, ow = out_im.shape[:2]

    Tinv = np.linalg.inv(T)
    for i in range(oh):
        for j in range(ow):
            coords = np.dot(Tinv, [i,j,1])

            src_i = coords[0] / coords[2]
            src_j = coords[1] / coords[2]
            out_im[i,j,:] = bilerp(im, src_i, src_j)


def bilerp(im, i, j):
    """ bilinear interpolation: return interpolated pixel
    value at floating-point coordinates i,j in im. pixels
    outside im's bounds are assumed to be 0 """
    h, w = im.shape[:2]

    # integer pixel location up and to the left of i,j
    itop, jleft = int(np.floor(i)), int(np.floor(j))

    # return black if out of bounds
    if itop > h-2 or itop < 0 or jleft > w-2 or j < 0:
        return np.zeros(3)

    # calculate weights for each pixel
    tl_wt = (itop + 1 - i) * (jleft + 1 - j) # top left
    bl_wt = (i - itop)     * (jleft + 1 - j) # bottom left
    tr_wt = (itop + 1 - i) * (j - jleft) # top right
    br_wt = (i - itop)     * (j - jleft)# bottom right

    # accumulate weighted pixel values
    pixel  = tl_wt * im[itop, jleft,:]
    pixel += bl_wt * im[itop+1, jleft,:]
    pixel += tr_wt * im[itop, jleft+1,:]
    pixel += br_wt * im[itop+1, jleft+1,:]

    return pixel


if __name__ == "__main__":

    van = cv2.imread('van.png').astype(np.float32) / 255.0

    h, w = van.shape[:2]

    # translate center of image to (0,0)
    ti, tj = h//2, w//2
    T = np.array([
        [1, 0, ti],
        [0, 1, tj],
        [0, 0, 1]], dtype=np.float32)

    # rotate 45 degrees around (0,0)
    th = np.pi / 4
    R45 = np.array([
        [cos(th), -sin(th), 0],
        [sin(th),  cos(th), 0],
        [0,        0,       1]])


    # rotate 45 degrees around the center point
    Rmid = T.dot(R45.dot(np.linalg.inv(T)))

    h, w = van.shape[:2]
    out_im = np.zeros((h, w, 3))

    warp(van, Rmid, out_im)

    cv2.imwrite('van_warp.png', (out_im * 255.0).astype(np.float32))

